-- Substitute approximations for trig functions and others.  Do before
-- CSE, since it's likely to produce common sub-expressions.
-- 
-- TODO: Expand this phase to be a mapping from an architecture-neutral
-- language into an architecture-specific one.  For instance, the neutral
-- language would have pow, while the specific one would have rsq, rcp,
-- 2**x.  The neutral one would have sin & cos that work on arbitrary
-- reals, while the specific one might not have them at all or need
-- argument normalization.  In the latter case, it's important that we
-- have access to interval information so we can choose the approximation.

module Approx where

import Memo

import DExp
import qualified SmartDExp as SD
import qualified TExp as T

approx :: Exp -> Exp
approx e = approxM e
 where
   approxM = memo approxN
   approxN (Apply Sin [u])  = approxM (unty1 T.sinApproxDouble u)
   approxN (Apply Cos [u])  = approxM (unty1 T.cosApproxDouble u)
   approxN (Apply Frac [u]) = approxM (fracApprox u)
   approxN (Apply op args)  = reOp op (map approxM args)
   approxN e = e

-- Should we use SD.applyOpt or Apply in approx.  The former would try
-- optimizations after substituting approximations.  I did a few
-- experiments, and found the results inconclusive.  Sometimes better,
-- sometimes worse, nevery by much.

reOp = SD.applyOpt
       -- Apply

unty1 :: (T.Exp a -> T.Exp b) -> (Exp -> Exp)
unty1 tyf = T.unE . tyf . T.mkE

fracApprox :: Exp -> Exp
fracApprox e = SD.vecY (Apply Expp [e])

-- TODO: fracApprox doesn't currently optimize well.  The expp stores its
-- result in a 4-tuple, and the vecY uses just one part.  Better would be
-- to mask the expp instruction.  I think to do this optimization, we'd
-- want to know that there is no other use of the expp result vector that
-- accesses other components.  Another sharing-dependent rewrite.
