-- Translation to C/C#

module CTrans (csnorm) where

import Types
import DExp
import PP
import PPDExp (ppExp)
-- import qualified TExp as T

-- C# normal form.  In this normal form, we can have Lets at the top and
-- at the top of RHSs of bindings in Let expressions, and as the bodies of
-- Let expressions.  On second thought, don't allow them in binding RHSs
-- either, so as to make code generation simpler.
type CSNF = Exp

-- Normalize for C# generation.  Assumption: there are no free occurrences
-- of any let-bound variable, so we don't have to do renaming.

csnorm :: Exp -> CSNF

-- In an application, arguments can be wrapped in more than one layer of
-- Lets.  Peel off one layer at a time until there are no more, using
-- normArgs.

csnorm (Apply op args) = normArgs (Apply op) (map csnorm args)

csnorm (Let binds body) =
  normArgs (\ exps' -> Let (zip vars exps') (csnorm body)) exps
 where
   (vars,exps) = unzip binds

-- Simpler form that allows Lets in binding RHSs.
-- csnorm (Let binds body) = Let (map (onSnd csnorm) binds) (csnorm body)

csnorm e = e

-- cnorm :: T.Exp a -> T.Exp a
-- cnorm = T.typ1 csnorm

normArgs :: ([CSNF] -> CSNF) -> [CSNF] -> CSNF
normArgs f args | null binds = f args'
                | otherwise  = Let binds (normArgs f args')
 where
   (bindss,args') = unzip (map stripLet args)
   binds          = concat bindss


stripLet :: Exp -> ([(Id,Exp)],Exp)
stripLet (Let binds body) = (binds,body)
stripLet e                = ([],e)


{-
---- Now the actual C# code generation.

ppCs :: Exp -> PPDoc
ppCs e = stmtBlock 4 (vcat (map ppDecl decls) $$ ppExpForCs body)
 where
   (decls,body) = stripDecls (csnorm e)

-- Strip off all surrounding declarations, concatenating the bindings.
stripDecls :: CSNF -> ([(Id,Exp)],Exp)
stripDecls (Let binds body) = (binds ++ binds', body')
 where
   (binds',body') = stripDecls body

stripDecls e = ([],e)

---- Pretty print expressions for C#
-}
