-- Function overloadings.  Needed for the explicit parameter version of
-- Shader.hs.

module Function where

---- Now for functions.  They apply to curried functions as well as non-curried.

-- Some handy "lifting" functionals:
lift0 = const
lift1 = (.)
lift2 h f1 f2 x = h (f1 x) (f2 x)
lift3 h f1 f2 f3 x = h (f1 x) (f2 x) (f3 x)

-- Function-level pairing
pairF :: (c -> a) -> (c -> b) -> (c -> (a,b))
pairF = lift2 (,)


noOv meth = error $ "No function overloading for " ++ meth

-- Eq & Show are prerequisites for Num, so they need to be faked here
instance Eq (a->b) where
  (==) = noOv "(==)"
  (/=) = noOv "(/=)"

instance Ord b => Ord (a->b) where
  min = lift2 min
  max = lift2 max

instance Show (a->b) where
  show      = noOv "show"
  showsPrec = noOv "showsPrec"
  showList  = noOv "showList"

instance Num b => Num (a->b) where
  (+)         = lift2 (+)
  (*)         = lift2 (*)
  negate      = lift1 negate
  fromInteger = lift0 . fromInteger

instance Fractional b => Fractional (a->b) where
  recip        = lift1 recip
  fromRational = lift0 . fromRational

instance Floating b => Floating (a->b) where
  sqrt = lift1 sqrt
  exp  = lift1 exp
  log  = lift1 log


class    Liftable k f where lift            :: k -> f
instance Liftable a a where lift            = id
instance Liftable b (a->b) where lift       = const . lift
instance Liftable c (a->b->c) where lift    = const . lift
instance Liftable d (a->b->c->d) where lift = const . lift


{-
-- Is there a general recursive definition?  The following is declared overlapping
-- instance Liftable a c => Liftable a (b->c) where lift = const . lift

-- Some tests
test f = f 3 "foodles"

type T a = Int -> String -> a

f1,f2 :: T Int
f1 n str = n + length str               -- 10
f2 n str = 5 * n - length str           -- 8

f3 = f1 + f2                            -- 18
f4 = f1 * f2 + 7                        -- 87

-}