-- Generate 4^4 overloaded swizzling functions.  
-- Run "main" in this module to write SwizzleFuns.hs.

module GenSwiz where

import List (intersperse,nub)
import Maybe (fromJust)
import Char (toLower)
import TExp

swizCode :: [Int] -> String
swizCode ixs =
  fname ++ " :: " ++ tupstr (nub (map context coords)) ++
  " => Exp a -> Exp " ++ tupstr (map ((:[]). toLower) coords) ++ "\n" ++
  fname ++ " = " ++ "swiz" ++ show n ++ " " ++
  tupstr (map accessor coords) ++ "\n"
 where
   fname      = "swiz" ++ coords
   coords     = map (assoc (zip [0..] "XYZW")) ixs
   context  c = "Vec" ++ [c] ++ " a " ++ [toLower c]
   accessor c = "vec" ++ [c]
   n          = length ixs

   tupstr :: [String] -> String
   tupstr strs = "(" ++ concat (intersperse "," strs) ++ ")"

   assoc keyVals key = fromJust (lookup key keyVals)

swizCodeAll = unlines $
              [ "-- Generated by GenSwiz.hs"
              , "module SwizzleFuns where"
              , "import TExp"
              , ""
              ]
              ++
              map swizCode (tail (allSpecs 4))

allSpecs n = if n < 0 then []
             else [] : [c:spec | spec <- sp', c <- [0..3]]
  where sp' = allSpecs (n-1)

main :: IO ()
main = do writeFile "SwizzleFuns.hs" swizCodeAll
          putStrLn "Wrote SwizzleFuns.hs"
