
-- Simple camera, with position and look & up vectors.  Doesn't crop or
-- discretize.
data Camera = 
  Camera { cameraPos  :: Point3
         , cameraLook :: Vector3
         , cameraUp   :: Vector3
         }



codegen (Let v e1 e2) = (nr,gen)
 where
   (nr1,gen1) = codegen e1
   (nr2,gen2) = codegen e2
   nr = max nr1 (1+nr2)   -- reserve v during e2
   gen r env pool = gen1 r1 env pool ++
                    gen2 r (extendEnv env v r) pool'
     where
       (r1,pool') = getFree pool e1 env

codegen (Apply op args) = (nr,gen)
 where
   (nrs,gens) = unzip (map codegen args)
   indices = order (map negate nrs)
   
   -- reordered
   nrs'  = resort indices nrs
   args' = resort indices args
   gens' = resort indices gens
   
   needs' = map needSave args'
   -- How many registers to save at each argument evaluation
   savess' = --trace ("nrs' == " ++ show nrs' ++ "\n") $
             --trace ("args' == " ++ show args' ++ "\n") $
             --trace ("needs' == " ++ show needs' ++ "\n") $
             0: scanl1 (+) needs'
   
   nr = --trace ("savess' == " ++ show savess' ++ "\n") $
        maximum [ nr' + saves' | (nr',saves') <- zip nrs' savess' ]
   
   gen r env pool = asm ++
                    [PrimOp r op (unsort indices regs)]
    where
     (asm,regs) = g args' gens' pool
     g [] [] _ = ([],[])
     g (e1:es) (gen1:gens') pool = (asm1 ++ asms,r1:regs)
       where (r1,pool') = getFree pool e1 env
             asm1 = gen1 r1 env pool
             (asms,regs) = g es gens' pool'


needSave :: Exp -> Int
needSave (Var _) = 0
needSave _       = 1


order :: Ord a => [a] -> [Int]
order = map fst . sortBy (comp2 snd) . zip [0..]

comp2 f a b = f a `compare` f b

unsort, resort :: [Int] -> [a] -> [a]

unsort indices = map snd . sortBy (comp2 fst) . zip indices

resort indices as = map (as !!) indices



parseRegOut :: String -> Int
parseRegOut "oD0" = 0
parseRegOut "oFog" = 2
parseRegOut "oPos" = 3
parseRegOut "oPts" = 4
parseRegOut ('o':'T':'e':'x':numStr) = 5 + read numStr



-- Tweak argument list to get the effect of negating the Op's result.
-- This interface may prove inadequate for some operators.  If so,
-- revise.
negateOp :: Op -> [Source] -> [Source]

negateOp Rcp = negateAll
negateOp Add = negateAll
negateOp Mul = negateHead
negateOp Mad = negateTail
negateOp Dp4 = negateTail

-- Some common patterns
negateAll, negateHead, negateTail :: [Source] -> [Source]
negateAll = map negateSrc
negateHead (src:srcs) = negateSrc src : srcs
negateTail (src:srcs) = src : negateAll srcs

ifF :: Bool -> (a -> a) -> (a -> a)
ifF False f = id
ifF True  f = f


-------------------------------



partOf :: Part -> Quad a -> a
partOf X (x,_,_,_) = x
partOf Y (_,y,_,_) = y
partOf Z (_,_,z,_) = z
partOf W (_,_,_,w) = w



arity :: Op -> Int
arity Rcp = 1
arity Add = 2
arity Mul = 2
arity Dp4 = 2
arity Mad = 3


-- Tweak argument list to get the effect of negating the Op's result.
-- This interface may prove inadequate for some operators.  If so,
-- revise.
negateOp :: Op -> [Arg] -> [Arg]

negateOp Add = negateAll   --  -(a+b)   == -a + -b
negateOp Mad = negateTail  --  -(a*b+c) == a * -b + c
negateOp Mul = negateHead  --  -(a*b)   == -a * b
negateOp Rcp = negateHead  --  - 1/a    == 1/-a
negateOp Dp3 = negateHead  --  -(a<.>b) == -a <.> b
negateOp Dp4 = negateHead  --  -(a<.>b) == -a <.> b




class Indexed ix coll where
  elt :: ix -> col a -> a

data Part2 = X2 | Y2
data Part3 = X3 | Y3 | Z3
data Part4 = X4 | Y4 | Z4 | W4

instance Indexed Part2 (a,a) where
  elt X2 (x,y) = x
  elt Y2 (x,y) = y

instance Indexed Part3 (a,a,a) where
  elt X3 (x,y,z) = x
  elt Y3 (x,y,z) = y
  elt Z3 (x,y,z) = z

instance Indexed Part4 (a,a,a,a) where
  elt X4 (x,y,z,w) = x
  elt Y4 (x,y,z,w) = y
  elt Z4 (x,y,z,w) = z
  elt W4 (x,y,z,w) = w

swizzle :: (Indexed ixD collD,Indexed ixR collR) =>
           ...MUMBLE...

overload x,y,z




-- In computer graphics, good things come in fours.
type Quad a  = (a,a,a,a)

quadToList :: Quad a -> [a]
quadToList (x,y,z,w) = [x,y,z,w]

listToQuad :: [a] -> Quad a
listToQuad [x,y,z,w] = (x,y,z,w)

mapQuad :: (a->b) -> Quad a -> Quad b
mapQuad f (x,y,z,w) = (f x,f y,f z,f w)

type Mask = Quad Part

data Part = X | Y | Z | W  deriving (Show,Eq,Ord)




destPart :: Part -> Dest -> Dest
destPart part (Dest reg mask)
   | pmask `subMask` mask = Dest reg pmask
 where
   pmask = partMask part

   partMask :: Part -> Mask
   partMask part = mapQuad (==part) idSwizzle

   subMask :: Mask -> Mask -> Bool
   subMask m1 m2 = and (zipWith implies (quadToList m1) (quadToList m2))

destPart part dest =
  error $ "destPart: " ++ show part ++ " (" ++ show dest ++ ")"

-- negateSrc :: Source -> Source
-- negateSrc (Source ns reg) = Source (negateNS ns) reg

idMask :: Mask
idMask = (True,True,True,True)

allParts :: [Part]
allParts = quadToList idSwizzle


   match :: Index -> Bool -> Index -> Bool
   match refIx inDest swizIx = inDest `implies` (swizIx==refIx)

implies :: Bool -> Bool -> Bool
a `implies` b = not a || b




-- idLayout :: Layout
-- idLayout = quadToList idSwizzle

-- TODO: 
--  layoutMask :: Layout -> Mask
--  reLayout :: Layout -> Layout -> Swizzle



padLeftTo :: Int -> a -> [a] -> [a]
padLeftTo n x xs = replicate (n - length xs) x ++ xs


{-

+ What happens to code generation, and particularly argument ordering,
  when there are registers of different types?  Instead of just one number
  for the minimum registers consumed during expression evaluation, there
  is one per type.  Maybe more complex than that.  For now, I'm counting
  the total number of floats.  There will be integer registers as well.
  In the case that the resource uses are strictly comparable (same
  relationship for every register type), we use the same algorithm as now,
  but when they're not, we try both.  Ask Chris Fraser when I need to deal
  with that issue, since it must be a standard extension to the code
  generation algorithm.

+ Negation and swizzling at the top level and as binding RHSs generates
  explicit Mov instructions, which could be avoided by using distributive
  properties of the PrimOps.  See NegateOp in Junk.hs.

+ Vertex shaders define more than one result value.  How to
  accommodate?  Ideas:

  - Generate code for a tuple or list of expressions, given a tuple or
    list of registers.

  - Maybe simply a new "operator" that assigns its arguments to a
    given sequence of output registers indicated by number:

      data Op = ... | Output [Int]

    A suboptimal aspect of this simple formulation is that the code
    generator thinks that the separate arguments to an output operator
    must coexist, which unnecessarily reduces the number of available
    registers.

  - More simply, add an explicit multi-way assignment operator to the
    source language, assigning to output vertex registers.

      data Exp = ... | Output [(Int,Exp)]

    When I layer on a typed layer, then Output will be an Action, so
    it will not compose with values.  Don't add a sequencing
    operation, so an Action expression will necessarily be some
    enclosing Let bindings around an Output expression.

    Compilation works as currently, with a simple additional case for
    Output.  Unlike Let, Output doesn't need to preserve registers
    between argument evaluations.

  - Have a VertexOut data type with labeled fields and defaults that allow
    the compiler to know what's been set and what hasn't.  (Could match
    the vertex shader output register defaults, or could be a special
    syntactic "undefined" form, which might come in handly elsewhere.)

-}


eVal :: Exp a -> a
eVal = undefined



instance ExpType Float where expType _ = scalar
instance ExpType Bool where expType _ = scalar

instance ToFloat a => ExpType Float where expType _ = 1

instance (ToFloat a, ToFloat b) => ExpType (a,b) where
  expType _ = 2

instance (ToFloat a, ToFloat b, ToFloat c) => ExpType (a,b,c) where
  expType _ = 3

instance (ToFloat a, ToFloat b) => ExpType (a,b) where
  expType _ = 4




ifZ :: Exp -> Exp -> Exp -> Exp

-- Merge nested conditionals that have the same condition.  The
-- combination of "if splitting" done by the (a,b) and Color Syntactic
-- instances, and the if lifting done by opt<n>, yields many mergeable
-- conditionals.
-- ifZ c (If c' a' b') b | c ==% c' = ifE c a' b
-- ifZ c a (If c' a' b') | c ==% c' = ifE c a b'
ifZ _ a b             | a ==% b  = a
-- ifZ c a b                        = If c (assume c a) (assume (notE c) b)

-- The first two clauses appear to be redundant with the last.  ToDo:
-- remove them and check that we lose no reductions.

{-

-- Could use this version instead of the previous one.  This
-- implementation collects all conditionals, but one is probably
-- sufficient, since simplification is done bottom-up.
ifZ c a b                       = assume [] (If c a b)


-- Simplify a sub-expression given a list of boolean expressions known to
-- be true.  Currently, only contained conditionals are simplified.

assume :: [Exp] -> Exp -> Exp
assume asms (If c a b) 
  |      c `elem` asms = assume asms a
  | notE c `elem` asms = assume asms b
  | otherwise          = If c (assume (     c : asms) a)
                              (assume (notE c : asms) b)

-- Could recursively simplify non-if sub-expressions here.  If so, memoize.
assume assume e = e

-- This implementation uses just a single condition.
assume :: Exp -> Exp -> Exp
assume asm = assumeRec
 where
   assumeRec (If c a b) 
    |      c ==% asm  = assumeRec a
    | notE c ==% asm  = assumeRec b
    | otherwise       = If c (assumeRec a) (assumeRec b)

   assumeRec e = e
-}

select :: Int -> Exp -> Exp
select i = apply1 (Swiz [i])

vector :: [Exp] -> Exp
vector = MkVec

-- TODO: optimize vector.  In particular, all arguments may be swizzles of
-- the same expression.



-- The general function, doing various general optimizations.
ifE :: Exp -> Exp -> Exp -> Exp

ifE (Not c) a b = ifE c b a

ifE c a b | a ==% b     = a

ifE (LitVec cs) a b = vector (zipWith select [0..] (map pick cs))
 where
   pick 0.0 = b
   pick 1.0 = a

-- ifE (If c d e)      a b = ifE c (ifE d a b) (ifE e a b)

ifE c a b               = ifZ c a b

-- Is the if/if rule a good idea?  It certainly bloats.  I did it because
-- a and b often have c-based conditionals as well.  Moving them under the
-- "if c" gives them a chance to simplify.  Experimental.



-- Dissect expression as if NegateSwizzle
asNegSwiz :: D.Exp -> (NegSwiz, D.Exp)
asNegSwiz dexp = gather False (idSwizzle (typeOf dexp)) dexp
 where
   gather negB swizB (Apply Negate [e]) =
     gather (not negB) ns e
   gather negB swizB (Apply (Swizzle swizA) [e]) =
     gather negB (swizB `compSwiz` swizA) e
   gather negB swizB e = (NegSwiz negB swizB, e)


normalize (D.Let dbinds dbody) =
  do let (vars,dexps) = unzip dbinds
     nexps <- mapM normalize dexps
     body <- normalize dbody
     return $ Let (zip vars nexps) body


class Simd a where simd :: (Float->Float) -> (a->a)

instance Simd Float where simd = id

instance Simd (Float,Float) where
  simd f (x,y) = (f x, f y)

instance Simd (Float,Float,Float) where
  simd f (x,y,z) = (f x, f y, f z)

instance Simd (Float,Float,Float,Float) where
  simd f (x,y,z,w) = (f x, f y, f z, f w)




   try i reg = fmap constArg (mapM (`elemIndex` reg) vec)   -- Maybe monad
     where constArg ixs = Arg (NegSwiz False ixs)
                              (constVarName i)
                              (size2type (length vec))



addRegState :: Int -> [Float] -> RegState -> RegState
addRegState reg vec = add 0 []
 where
   -- Reverse and accumulate state as we go.
   add _ _ [] = error "addRegState: Ran out of constant space"
   add i revRS | i == reg =  MUMBLE



----------------------------------------------------------------------


---- Constants

-- TODO: split this section into its own module

-- State of a bank of registers.  The values contained in each.
type RegState = [Vector]

initRegs :: Int -> RegState
initRegs n = take n (repeat [])

addRegF :: Vector -> RegState -> RegState
addRegF vec regs | null r    = error "addReg: Ran out of constant space"
                 | otherwise = dontFits ++ (mergeWith fits : rest)
 where
   (dontFits, r@(fits:rest)) = break fit regs
   fit reg                   = length (mergeWith reg) <= 4
   mergeWith reg             = nub (reg ++ vec)

-- TODO: elim redundant computation of mergeWith above.

-- TODO: minimize register consumption better.  Get as much reuse as we
-- can in addReg.  

-- TODO: consider global minimization rather than a greedy algorithm.
-- Could pack up as a circular function that takes in a fixed collection
-- of register values and uses it to replace its constants, returning the
-- set of constants found.  Put in circular dependence with the
-- optimization function that maps a set of constants to collection of
-- register values.


-- A monad for finding and extracting constants.
newtype FC a = FindC (RegState -> (a,RegState))

instance Monad FC where
  FindC fc >>= g = FindC $ \ st ->
                             let (a,st')     = fc st
                                 (FindC gfc) = g a
                             in
                                 gfc st'
  return a = FindC $ \ st -> (a,st)

instance Functor FC where
    fmap f m = m >>= return.f

getRegs :: FC RegState
getRegs = FindC $ \ st -> (st,st)

setRegs :: RegState -> FC ()
setRegs st' = FindC $ \ _ -> ((),st')

addReg vec = do regs <- getRegs
                setRegs (addRegF vec regs)

makeConst :: Vector -> FC Arg
makeConst vec =
  -- Look for an existing constant that contains all the floats in vec,
  -- possibly in a different order, possibly using the same float more
  -- than once, and possibly all negated.  If there isn't one, allocate
  -- more space.  In any case, yield an arg.
  do regs <- getRegs
     maybe (newConst vec)
           return
           (findConst vec regs)
      
newConst :: Vector -> FC Arg
newConst vec = do addReg vec
                  -- Now we can surely find vec.
                  regs' <- getRegs
                  return $ fromJust (findConst vec regs')

-- TODO: make findConst an FC (Maybe Arg) & redo makeConst def.

findConst :: Vector -> RegState -> Maybe Arg
findConst vec = loop 0
 where
   loop _ [] = Nothing
   loop i (reg:regs') = try i reg `mplus` loop (i+1) regs'
   
   try :: Int -> Vector -> Maybe Arg
   try i reg =
     do ixs <- mapM (`elemIndex` reg) vec   -- Maybe monad
        return $ Arg (NegSwiz False ixs)
                     (constVarName i)
                     (size2type (length vec))
-- Test:
fc1 = findConst [3,3,5] [[3.0,5.0,4.0]]

constVarName :: Int -> String
constVarName n = "_const_" ++ show n

runFC :: Int -> FC a -> (a, RegState)
runFC n (FindC fc) = fc (initRegs n)

-- TODO: Accommodate the need for variables that are constant within a
-- mesh, which will be set outside of the generated shader.



-- Tests:
t1 = normalize $ LitVec [5,6]
t2 = normalize $ LitVec [5,6,5]
t4 = normalize $ apply2 Slt (Var "a" (Vec 2)) (LitVec [5,6])



---- Typing

typeOfArg :: Arg -> Type
typeOfArg (Arg _ _ ty) = ty

---- Convenient expression-building functions.

var :: Id -> Type -> CNF
var = Var


----------------------------------------------------------------------

-- From CodeGen.hs

codegen (Var v _) = (0, gen)
 where
   -- Will this nopMov optimization ever fire?  Put in later when the need
   -- surfaces.  Then we can test it.
   gen dest swiz env _
     {- | nopMov dest src = []
        | otherwise -}  = [Mov dest (swizzleSrc swiz src)]
     where src = lookupEnv env v

codegen (NegateSwizzle nsA (Var v _)) = (0, gen)
 where
   gen dest swizB env _ =
     [Mov dest (negSwizLookup env (swizzleNS swizB nsA) v)]

codegen (NegateSwizzle ns e) =
  -- This case should be rare -- only at the top level or from the first
  -- Apply case below.
  codegen (Let [(v,e)] (NegateSwizzle ns (Var v (typeOf e))))
 where
   v = "_ns0"

-- Make sure that we can eliminate swizzle applied to Apply here.  At last
-- resort, wrap a NegateSwizzle.
codegen e@(Apply op args) | not canOptSwiz =
  codegen (NegateSwizzle (idNegSwiz (typeOf e)) e)
 where
   v = "_ns0"
   canOptSwiz =    typeOf e == Vec4           -- guarantees identity layout
                || op `elem` simdOps          -- swizzle each arg
                || op `elem` scalarResultOps  -- swizzling is a no-op

----------------------------------------------------------------------

normalize e | ns /= idNegSwiz ty = normalize (apply1 Mov e)
 where
   ty      = typeOf e
   (ns,e') = asNegSwiz e

normalize e | ns /= idNegSwiz ty = 
  normalize (apply1 Mov (applyNS ns ty e'))
 where
   ty      = typeOf e
   (ns,e') = asNegSwiz e

normalize e | ns /= idNegSwiz ty = 
  apply1 Mov (applyNS ns ty (norm e'))
 where
   ty = typeOf e
   (ns,e') = asNegSwiz e
   -- After stripping off negates & swizzles, if we have a var, we're
   -- done.  Otherwise, normalize some more.
   norm e'@(Var _ _) = e'
   norm e'           = normalize e'


----------------------------------------------------------------------

expToTree :: Exp -> Tree Label
expToTree = memo mkTree
 where
   mkTree exp = Br label (map expToTree exps)
    where
      (label,exps) = dsExp exp

treeToGraph :: Tree a -> Graph a ()
treeToGraph = mkG empty
 where
   mkG tree@(Br label children) g = (if alreadyInG g (context & g'), node)
    where
      node = 
      context = ([],node,label, map (() `pair`) argNodes)
      (g',argNodes) =



toGraph = memo mkG -- Construct bottom up and memoize.
 where
   -- I'll need a little cheat to create the context to add, since it
   -- should depend on the graph as it's being built up, which would
   -- defeat memoization.  Do a little unsafe construction that increments
   -- an int ref value holding the next node number.  Rather like
   -- unsafePtrToInt.
   mkG ...



-- The main function is "share".  It finds shared nontrivial expressions
-- and hoists them out via appropriately placed lets.  The plan:
-- 
-- + Convert the expression into an explicit graph.
-- + Find the dominator set of each shared expression.
-- + Turn back into an expression.  For each shared non-trivial node,
--   place a binding at its immediate dominator and a corresponding
--   variable at its uses

----------------------------------------------------------------------

-- Place abstractions implied by node map
graphToExp :: EGraphN -> NodeMap [Node] -> Exp
graphToExp (egraph,n0) absMap = fromG emptyFM (allNames \\ egVars egraph) n0
 where
   fromG :: NodeMap Exp -> NameSupply -> Node -> Exp
   fromG env nameSup node =
     -- If this node is already abstracted as a variable, use that
     -- variable.  Otherwise, recursively convert children, make an
     -- expression, and wrap with bindings if any (if node is an
     -- abstraction site).
     fromMaybe (mkLets groups env nameSup) (lookupFM env node)
    where
      (Just (_,_,label,s),_) = match node egraph
      children               = map snd s
      -- Set of nodes to be abstracted here
      absNodes = fromMaybe [] (lookupFM absMap node)
      -- Stratify the set of nodes to be abstracted into a sequence of
      -- node sets, in which elements of later sets may depend on elements
      -- of earlier sets.  Make this stratification by performing
      -- breadth-first enumeration of the subgraph of egraph with nodes in
      -- absNodes, and coalescing elements with the same level number.
      groups = --ifF (not (null absNodes)) (ptrace "stratifyn") $
               stratifyn absNodes (subgraph (`elem` absNodes) egraph)
      mkLets [] env nameSup = mkExp label (map (fromG env nameSup) children)
      mkLets (group1:groups) env nameSup =
        let 
            exps = map (fromG env nameSup) group1
            (names,nameSup') = getNames (length group1) nameSup
            vars = zipWith Var names (map typeOf exps)
            env' = addListToFM env (zip group1 vars)
        in
            Let (zip names exps) (mkLets groups env' nameSup')
         

----------------------------------------------------------------------

-- Stratify a graph into bfs levels
stratifyn :: [Node] -> Graph a b -> [[Node]]
stratifyn ns = map (map fst) .                     -- just nodes
               groupBy (\ (_,i) (_,j) -> i==j) .   -- group by level
               --ifF (not (null ns)) (ptrace "leveln") .
               leveln (map (`pair` 1) ns)         -- bfs with level numbers

stratify :: Node -> Graph a b -> [[Node]]
stratify n = stratifyn [n]


----------------------------------------------------------------------

e19 = z + vr *^ z where z = evenTaylor vr

e20 = 3 * z + 4 * (vr *^ z) where z = evenTaylor vr

e21 = 3 <.> z + 4 <.> (vr *^ z) where z = evenTaylor vr

e22 = 3 <.> (vr *^ z) + 4 <.> z where z = evenTaylor vr
e23 = vec4 1 2 3 4 <.> (vr *^ z) + vec4 1 5 6 7 <.> z where z = evenTaylor vr

-- e24 CSEs the Taylor vector while e25 doesn't.
e24 = sinCoeffs <.> (vr *^ z) + cosCoeffs <.> z where z = evenTaylor vr

e25 = sinCoeffs <.> oddTaylor vr + cosCoeffs <.> evenTaylor vr
e26 = oddTaylor vr + evenTaylor vr      -- Doesn't

e27 = evenTaylor vr + evenTaylor vr     -- Doesn't
e28 = z + z where z = evenTaylor vr     -- Does

e29 = vec2 (vr^0) (vr^2) + vec2 (vr^0) (vr^2) -- doesn't

e30 = vec2 vr vq + vec2 vr vq           -- does

e31 = vec2 1 vq + vec2 1 vq             -- doesn't
e32 = z + z where z = vec2 1 vq         -- does

e33 = (1+vq) * (1+vq)                   -- doesn't

-- Oh, I know the problem.

-- See CSE.uniquifyExp comment
-- TODO: carefully copy literals, etc.

----------------------------------------------------------------------

negateE :: CNF -> CNF
negateE exp
  | negateCanOpt exp = negateOpt exp
  | otherwise = NegateSwizzle (NegSwiz True (idSwizzle (typeOf exp))) exp

-- Rewrite away the negation where possible.

negateOpt (NegateSwizzle ns e) = NegateSwizzle (negateNS ns) e

negateOpt (Apply Add args) = Apply Add (negateAll args)
negateOpt (Apply Max args) = Apply Min (negateAll args)
negateOpt (Apply Min args) = Apply Max (negateAll args)
negateOpt (Apply Mad args) = Apply Max (negateTail args)
negateOpt (Apply op args)
  | op `elem` [Mul,Rcp,Dp3,Dp4] = Apply op (negateHead args)

negateOpt (Vec exps) = Vec (map negateOpt exps)

-- For a let, push the negation into the body if it can be then optimized
-- away.  Otherwise, it can be harmful to do so, because more registers
-- will be tied up when the negation is performed.  (Actually, I think it
-- can only be harmful if the result register is an output reg.)
negateOpt (Let binds body) = Let binds (negateOpt body)

negateOpt e = error $ "negateOpt: couldn't optimize " ++ show e

-- Can negation be optimized away?  (Note: keep in sync with negateOpt def.)
negateCanOpt :: Exp -> Bool

negateCanOpt (Var _ _)           = False
negateCanOpt (NegateSwizzle _ _) = True
negateCanOpt (Apply op _)        = op `elem` [Add,Max,Min,Mad,Mul,Rcp,Dp3,Dp4]
negateCanOpt (Let _ body)        = negateCanOpt body
negateCanOpt (Vec exps)          = all negateCanOpt exps

-- TODO: reduce redundancy between negateE & negateCanOpt by introducing a

-- Helpers
negateArg :: Arg -> Arg
negateArg (Arg ns v ty) = Arg (negateNS ns) v ty

negateAll, negateHead, negateTail :: [Arg] -> [Arg]
negateAll = map negateArg
negateHead (arg:args) = negateArg arg : args
negateTail (arg:args) = arg : negateAll args



----------------------------------------------------------------------


negateOpt (Apply Negate [e]) = e
negateOpt (Apply Add exps)   = Apply Add (negateAll exps)
negateOpt (Apply Max exps)   = Apply Min (negateAll exps)
negateOpt (Apply Min exps)   = Apply Max (negateAll exps)
negateOpt (Apply Mad exps)   = Apply Max (negateTail exps)
negateOpt (Apply op exps)
  | op `elem` [Mul,Rcp,Dp3,Dp4] = Apply op (negateHead exps)

negateOpt (Apply MkVec exps) = Apply MkVec (negateAll exps)

-- For a let, push the negation into the body if it can be then optimized
-- away.  Otherwise, it can be harmful to do so, because more registers
-- will be tied up when the negation is performed.  (Actually, I think it
-- can only be harmful if the result register is an output reg.)
negateOpt (Let binds body) = Let binds (negateOpt body)

negateOpt e = error $ "negateOpt: couldn't optimize " ++ show e


-- Can negation be optimized away?  (Note: keep in sync with negateOpt def.)
negateCanOpt :: Exp -> Bool

negateCanOpt (Var _ _)           = False
negateCanOpt (Apply op _)        = op `elem` [Add,Max,Min,Mad,Mul,Rcp,Dp3,Dp4,Negate,MkVec]
negateCanOpt (Let _ body)        = negateCanOpt body




  , oT0 :: Exp, oT1 :: Exp, oT2 :: Exp, oT3 :: Exp
  , oT4 :: Exp, oT5 :: Exp, oT6 :: Exp, oT7 :: Exp



-- Colors & supporting values are defined in TExp at the typed expression level.

type Color  = (Float,Float,Float) -- RGB for now.  Assumed in [0,1]
type Colour = Color

rgb r g b = (r,g,b)

gray,grey :: Float -> Color
gray x = rgb x x x
grey   = gray

black :: Color
black = gray 0


----------------------------------------------------------------------

data VertexOut = VertexOut
  { oDiffuse  :: ColorE, oSpecular  :: ColorE
  , oFog :: FloatE                      -- fog, clamped to [0,1]
  , oPos :: Vec4E                       -- position
  , oPointSize :: FloatE                -- point sprite size
  , texCoords :: [Vec4E]                -- up to 8 texture coords
  }
  deriving (Eq,Show)

-- TODO: Figure out what type to make the texture coordinates.


-- Default value.  Convenient for building others, e.g.,
-- vertexOut{oDiffuse=e1,oPos=e2}
vertexOut :: VertexOut
vertexOut = VertexOut
  {oDiffuse=u,oSpecular=u,oFog=u,oPos=u,oPointSize=u,texCoords=[]}
 where u = undefinedE

type OutE = Exp VertexOut


----------------------------------------------------------------------


-- Well it's not quite that simple, because we don't want to abstract
-- negations or swizzlings, preferring to optimize away the negate or
-- swizzle at its uses.  So:

-- Abstract a node n iff n is an operator application other than a
-- negation or swizzle and n is either shared or "virtually" shared,
-- meaning that it's not a root and its parent is a negation or swizzle
-- that is either shared or virtually shared.

-- Oops, then extraction happens too low and so gets replicated.

shouldAbstract :: EGraphN -> Bool
shouldAbstract (g,n) = isOp label && not (isNegOrSwiz label) &&
                       (shared || virtuallyShared n)
 where
   label = lab n g
   shared = length (pre g n) > 1
   isNegOrSwiz (Op Negate)      = True
   isNegOrSwiz (Op (Swizzle _)) = True
   isNegOrSwiz _                = False
   isOp (Op _) = True
   isOp _      = False
   virtuallyShared m =
     singleParent && isNegOrSwiz (lab parent g) &&
     (length (pre g parent) > 1 || virtuallyShared parent)
    where
      parents      = pre g m
      singleParent = length parents == 1
      parent       = head parents

-- TODO: make shouldAbstract only be true for negations or swizzles when
-- the arguments should be abstracted.  Even then, it may be better to
-- abstract the unswizzled & unnegated terms instead, so the swizzling &
-- negation can be absorbed at uses.  No, leave that optimization up to
-- CNF or code generation.


----------------------------------------------------------------------

codegen (Apply MkVec exps) = (nr, gen)
 where
 -- Register use is max of the uses of the exps.  Use the float
 -- subregisters rather than allocating new ones.  Since we don't have
 -- to preserve registers, the evaluation order doesn't matter.
 (nrs,gens) = unzip (map codegen exps)
 nr         = maximum nrs
 gen dest swiz env pool = concat (zipWith genPart gens swiz)
  -- Generate for each exp, but put in swizzled positions.
  where
    genPart gen ix = gen (destPart ix dest) (idSwizzle scalar) env pool



codegen (Apply MkVec exps) = (nr, gen)
 where
 -- Register use is max of the uses of the exps.  Use the float
 -- subregisters rather than allocating new ones.  Since we don't have
 -- to preserve registers, the evaluation order doesn't matter.
 -- Oops!  That's not true.  We do have to preserve the registers that
 -- hold each component while computing the others.
 -- TODO: revisit and fix well.  Remove Pool.removeFromPool when I do this
 -- better.
 (nrs,gens) = unzip (map codegen exps)
 nr         = maximum nrs
 gen dest swiz env pool = concat (genParts gens swiz pool)
  where
    -- Generate for each exp, but put in swizzled positions.
    genPart gen part pool =
      gen (destPart part dest) (idSwizzle scalar) env pool
    -- genParts gens dests = zipWith genPart gens dests
    genParts [] _ pool = []
    genParts (gen:gens) (part:parts) pool =
      genPart gen part pool :
      genParts gens parts (removeFromPool regNum part pool)
    genParts _ _ _ = error "genParts"
    Dest (Reg _ regNum _) _ = dest



codegenMkVec :: [CNF] -> CodeGen
codegenMkVec exps = (nr, gen)
 where
 (nrs,gens) = unzip (map codegen exps)
 nr         = length exps + maximum nrs
 gen dest swiz env pool = concat (zipWith genPart gens swiz)
  -- Generate for each exp, but put in swizzled positions.
  where
    genPart gen ix = gen (destPart ix dest) (idSwizzle scalar) env pool


----------------------------------------------------------------------


simple :: TimeE -> Vec2E -> (ColorE,Vec4E)
simple t p = (col,p')
 where
   (u,v) = dsVec2 p
   col   = rgb u v 1
   p'    = vec4 u 0 v 1

triv :: TimeE -> Vec2E -> (ColorE,Vec4E)
triv t p = (gray 1,p')
 where
   (u,v) = dsVec2 p
   p' = vec4 u 0 v 1

triv2 :: (Mat4x4E,TimeE) -> Vec2E -> OutputE
triv2 (view,t) p = wrapColorPoint view (triv t p)

simple2 :: (Mat4x4E,TimeE) -> Vec2E -> OutputE
simple2 (view,t) p = wrapColorPoint view (simple t p)

----------------------------------------------------------------------

-- Throw in view/projection matrix and wrap output

wrapColorVec :: (m -> v -> (ColorE,Vec4E))
             -> ((Mat4x4E,m) -> v -> OutputE)
wrapColorVec f (viewProj,m) v =
  output [oDiffuse col, oPos (viewProj &* p)]
 where
   (col,p) = f m v

rippleOut (view,t) p = wrapColorVec view (ripple t p)

----------------------------------------------------------------------

-- No longer used.  Instead, avoidVars passed in.

-- Extract vars used in an expression graph
egVars :: EGraph -> [Id]
egVars = nub . catMaybes . map varName . nodeLabels
 where
   varName (Simple (Var name _)) = Just name
   varName _                     = Nothing


----------------------------------------------------------------------

---- Pointer-to-int conversion

ptrToNode :: a -> Node

{- BEGIN_HUGS_ONLY -}
primitive unsafePtrToInt "unsafePtrToInt" :: a -> Int
ptrToNode e = e `seq` unsafePtrToInt e
{- END_HUGS_ONLY -}

{- BEGIN_GHC_ONLY
ptrToNode e = e `seq` unsafePerformIO (makeStableName e)
   END_GHC_ONLY -}

-- TODO: The GHC definition of ptrToNode doesn't work, because its result
-- is a stable name, not an int (and so not a node).  How can I get a
-- unique integer?  (Note that hashStableName doesn't guarantee
-- uniqueness.)  This discrepancy between ints and stable names wasn't a
-- problem in Pan, because we were free to define Node as a stable name.
-- FGL has Node==Int wired in.


----------------------------------------------------------------------

addListToFM :: Ord a => FiniteMap a b -> [(a,b)] -> FiniteMap a b
addListToFM = foldl add
  where
    add fmap (a,b) = addToFM fmap a b



----------------------------------------------------------------------


-- memoT :: (Show a, Show b) => String -> (b -> a) -> (b -> a)
memoT msg = memo . ftrace' msg

memoT2 f = memo (memo  . f)

-- Replace equal expressions with eq ones.  Assume no lets.
uniquifyExp :: Exp -> Exp
uniquifyExp = uExp
 where
  uOp   = memoT "copyOp" copyOp
  uExp  = memoT "copyExp" copyExp
  uType = memoT "copyType" copyType
  -- uList = memo $ ftrace' "copyList" copyList
  uListFloat :: [Float] -> [Float]
  uListFloat = memoT "copyListFloat" (copyList uListFloat)
  uListExp :: [Exp] -> [Exp]
  uListExp = memoT "copyListExp" (copyList uListExp)
  uListChar :: [Char] -> [Char]
  uListChar = memoT "copyListChar" (copyList uListChar)
  uListInt :: [Int] -> [Int]
  uListInt = memoT "copyListInt" (copyList uListInt)

  copyOp (Swizzle swiz) = uSwizzle (uListInt swiz)
  copyOp (Out     regs) = uOut (uListInt regs)
  copyOp op             = op

  copyExp (LitVec vec)    = uLitVec (uListFloat (map uFloat vec))
  copyExp (Var v ty)      = uVar v (uType ty)
  copyExp (Apply op exps) = uApply (uOp op) (uListExp (map uExp exps))
  copyExp (Let _ _)       = error "uniquifyExp: found Let"
  copyExp Undefined       = Undefined

  copyType (Vec n) = uVec n

  -- copyList :: [a] -> [a]
  -- copyList []    = []
  -- copyList (h:t) = uCons h (uList t)

  copyList uList []    = []
  copyList uList (h:t) = uCons h (uList t)

  uLitVec  = memo  LitVec
  uVec     = memo  Vec
  uVar     = memo2 Var
  uApply   = memo2 Apply
  uCons    = memo2 (:)
  uSwizzle = memo  Swizzle
  uOut     = memo  Out

  -- How to uniquify a Float?  For now go via string.
  -- Is there an easier way?
  uFloat = memo read . uListChar . show

----------------------------------------------------------------------

  copyExp (Var v ty)      = uVar v (uType ty)
  copyExp (Apply op exps) = uApply (uOp op) (uListE (map uExp exps))
  copyExp (Let _ _)       = error "uniquifyExp: found Let"
  copyExp Undefined       = Undefined

  copyType (Vec n) = uVec n

  copyList []    = []
  copyList (h:t) = uCons h (uList t)

  uLitVec  = memo  LitVec
  uVec     = memo  Vec
  uVar     = memo2 var'   -- memo2 Var
  uApply   = memo2 apply' -- memo2 Apply
  uCons    = memo2 (:)
  uSwizzle = memo  Swizzle
  uOut     = memo  Out

  uListE = memoT "copyListE" copyListE'
  copyListE []    = nilE
  copyListE (h:t) = uCons h (uListE t)
  copyListE' l    = length (show l) `seq` copyListE l

  apply' = curry $ ftrace' "uncurry Apply" (uncurry Apply)

  var' = curry $ ftrace' "uncurry Var" (uncurry Var)

  -- How to uniquify a Float?  For now go via string.
  -- Is there an easier way?
  uFloat = memo read . uList . show

nilE = [] :: [Exp]


----------------------------------------------------------------------


class BoolTup bool

instance BoolTup Bool
instance BoolTup (Bool,Bool)
instance BoolTup (Bool,Bool,Bool)
instance BoolTup (Bool,Bool,Bool,Bool)

-- Boolean type corresponding to a given type.
class BoolTup bool => BoolType bool a | a -> bool where
  fromBool :: Exp bool -> Exp a

-- instance AsFloat a => BoolType Bool a

-- The line above is flagged inconsistent, although no tuples are declared
-- as AsFloat types.  Instead, list the AsFloat types here (redundantly).

-- If you add a BoolType type, make the corresponding addition for AsFloat.

instance BoolType Bool Float where
instance BoolType Bool Bool

instance (AsFloat a, AsFloat b) => BoolType (Bool,Bool) (a,b)
instance (AsFloat a, AsFloat b, AsFloat c) => BoolType (Bool,Bool,Bool) (a,b,c)
instance (AsFloat a, AsFloat b, AsFloat c, AsFloat d) => BoolType (Bool,Bool,Bool,Bool) (a,b,c,d)

(==*), (/=*) :: (BoolType bool a, Eq a) => Exp a -> Exp a -> Exp bool

(==*) = typ2 (SD.==*)
(/=*) = typ2 (SD./=*)

(<*), (<=*), (>=*), (>*) :: (BoolType bool a, Ord a)
                         => Exp a -> Exp a -> Exp bool
(<*)  = typ2 (SD.<*)
(<=*) = typ2 (SD.<=*)
(>=*) = typ2 (SD.>=*)
(>*)  = typ2 (SD.>*)

(&&*), (||*) :: BoolTup bool => Exp bool -> Exp bool -> Exp bool
(&&*) = typ2 (SD.&&*)
(||*) = typ2 (SD.||*)

notE :: BoolTup bool => Exp bool -> Exp bool
notE = typ1 SD.notE

ifE :: BoolType bool a => Exp bool -> Exp a -> Exp a -> Exp a
ifE cond a b = cond * a + notE cond * b

----------------------------------------------------------------------


instance VecX Vec1 where vecX = typ1 SD.vecX

instance VecX Vec2 where vecX = typ1 SD.vecX
instance VecY Vec2 where vecY = typ1 SD.vecY

instance VecX Vec3 where vecX = typ1 SD.vecX
instance VecY Vec3 where vecY = typ1 SD.vecY
instance VecZ Vec3 where vecZ = typ1 SD.vecZ

instance VecX Vec4 where vecX = typ1 SD.vecX
instance VecY Vec4 where vecY = typ1 SD.vecY
instance VecZ Vec4 where vecZ = typ1 SD.vecZ
instance VecW Vec4 where vecW = typ1 SD.vecW

instance VecX Vec3
instance VecY Vec3
instance VecZ Vec3

instance VecX Vec4
instance VecY Vec4
instance VecZ Vec4
instance VecW Vec4

----------------------------------------------------------------------


-- Oh, the list version couldn't work because it would have to know long
-- the list is, and we're only allowed to know the type.
instance AsVar a => AsVar [a] where
  asVar xs = do (xs',envs) <- fmap unzip (mapM asVar xs)
                return $ (xs', concat envs)

-- Oops! lights are functions.
instance AsVar View where
  asVar ~(View amb eye lights) = do (a,aenv) <- asVar amb
                                    (b,benv) <- asVar eye
                                    (c,cenv) <- asVar lights
                                    return $ (View a b c, aenv++benv++cenv)


----------------------------------------------------------------------

instance (AsVar a, AsVar b, AsVar c, AsVar d) => AsVar (a,b,c,d) where
  asVar ~(a,b,c,d) = do (a,aenv,adecls) <- asVar a
                        (b,benv,bdecls) <- asVar b
                        (c,cenv,cdecls) <- asVar c
                        (d,denv,ddecls) <- asVar d
                        return $ ( (a,b,c,d)
                                 , aenv++benv++cenv++denv
                                 , adecls++bdecls++cdecls++ddecls)

instance (AsVar a, AsVar b, AsVar c, AsVar d) => AsVar (a,b,c,d) where
  asVar ~(a,b,c,d) = do (a,aenv,adecls) <- asVar a
                        (b,benv,bdecls) <- asVar b
                        (c,cenv,cdecls) <- asVar c
                        (d,denv,ddecls) <- asVar d
                        return $ ( (a,b,c,d)
                                 , aenv++benv++cenv++denv
                                 , adecls++bdecls++cdecls++ddecls)

instance AsVar Mat4x4E where
  asVar ~(Mat4x4E a b c d) = do (a,aenv,adecls) <- asVar a
                                (b,benv,bdecls) <- asVar b
                                (c,cenv,cdecls) <- asVar c
                                (d,denv,ddecls) <- asVar d
                                return $ ( (a,b,c,d)
                                         , aenv++benv++cenv++denv
                                         , adecls++bdecls++cdecls++ddecls)


----------------------------------------------------------------------



setterCode :: [Vector] -> SourceEnv -> [PPDoc]
setterCode constVecs uniformEnv =
  zipWith setLiteral [0..] constVecs ++
  map (setUniform.snd) uniformEnv
 where
   setLiteral i vec = ppCall setVertexConst [ppInt i, newArr "float" vec]
   setUniform _ = text ("setUniform not yet implemented")
   -- setUniform (Source (NegSwiz RegConst regNum _) ...)

newArr :: String -> Vector -> PPDoc
newArr ty vec = text ("new " ++ ty ++ "[]{") <> vals  <> text "}"
 where
   vals = hcat (intersperse comma (map ppFloat vec))

-- Wrong:
uniformFS (var, Source (NegSwiz _ layout) (Reg RegConst regNum _)) =
  withReg regNum (map (FSVar var (Vec (length layout))) layout)


----------------------------------------------------------------------


{-
Hmmm... I don't think the sphere case comes out correctly.  Without
normalization, testN sphere yields the following:

  let
      a = cos u
      b = sin u
      c = sin v
  in 
   vec3 (c * c * a)
        (let
             d = cos v
         in 
          c * a * a * d + b * d * c * b)
        (c * b * c)

Note that the body of the "d" let could be factored into (a^2+b^2)*(c*d),
which equals c*d.  Note also that the X normal component is sin^2 v * cos
u, which I think is wrong.  Maybe normalization would fix it.

-}

{-
type SurfP a = R2 -> (R3,a)

-- Add normal vector
withNormal :: SurfP a -> SurfP (R3,a)
withNormal surf = \ p2 -> let (p3,a) = surf p2 in
                          (p3, (surfNorm p2, a))
 where
   surfNorm = normal surf
-}

----------------------------------------------------------------------


llet :: AsVector a => Id -> Exp a -> (Exp a -> Exp b)  -> Exp b
llet name (E val) fun = E (D.Let [(name,val)] (unE (fun (var name))))


-- Experimental (and not working): rename ifE above to ifE', and define a
-- more flexible version that allows scalar bools to be promoted into
-- vectors of bools.

class Coerce a b where coerce :: a -> b

-- instance Coerce a a where coerce = id
-- instance FromScalar s v => Coerce s v where coerce = fromScalar

-- This one can expand scalar bools into vector bools
ifE :: (BoolType bools a, Coerce bool bools)
    => Exp bool -> Exp a -> Exp a -> Exp a
ifE c = ifE' (coerce c)



type Sel a = a -> FloatE

swiz1 :: (Sel a) -> a -> Vec1E
swiz1 s1 vec = vec1 (s1 vec)

swiz2 :: (Sel a, Sel a) -> a -> Vec2E
swiz2 (s1,s2) vec = vec2 (s1 vec) (s2 vec)

swiz3 :: (Sel a, Sel a, Sel a) -> a -> Vec3E
swiz3 (s1,s2,s3) vec = vec3 (s1 vec) (s2 vec) (s3 vec)

swiz4 :: (Sel a, Sel a, Sel a, Sel a) -> a -> Vec4E
swiz4 (s1,s2,s3,s4) vec = vec4 (s1 vec) (s2 vec) (s3 vec) (s4 vec)



vec1 :: FloatE -> Vec1E
vec1 = id

vec2 :: FloatE -> FloatE -> Vec2E
vec2 = typ2 SD.vec2

vec3 :: FloatE -> FloatE -> FloatE -> Vec3E
vec3 = typ3 SD.vec3

vec4 :: FloatE -> FloatE -> FloatE -> FloatE -> Vec4E
vec4 = typ4 SD.vec4


instance AsVector a => AsVar (Exp a) where
  asVar e = do name <- getName
               return (var name, [(name,expType e)])

               
staged :: (AsVar uniform, AsVar varying)
       => (uniform -> varying -> Exp o)
       -> (Exp o, TypeEnv, TypeEnv)
staged f = runNamer allNames $
           do (withUniform,uniformEnv) <- unFunctionVM f
              (withVarying,varyingEnv) <- unFunctionVM withUniform
              return (withVarying, uniformEnv, varyingEnv)
               

----------------------------------------------------------------------


instance VectorOf FloatE Vec2E where
  s *^ vec = fromScalar s * vec
  u <.> v  = extend u <.> extend v
             where
               extend vec = vec3 (vecX vec) (vecY vec) (0::FloatE)




instance AsVar (String,String) Vec2E where
  asVar (xv,yv) vec = (env, vec2 x y) where
                        (env, (x,y)) = asVar (xv,yv) (coords vec)




noMatrices :: TypeEnv -> TypeEnv
noMatrices = concatMap noMatrix
 where
   noMatrix :: (Id,Type) -> TypeEnv
   noMatrix (name, Matrix cols rows) =
     [(name ++ "_" ++ show i, Vec cols) | i <- [1 .. rows]]
     -- Equivalently,
     -- map (\ i -> (name ++ "_" ++ show i, Vec cols)) [1 .. rows]
     -- map ((`pair` (Vec cols)) . (((name ++ "_") ++) . show)) [1 .. rows]


----------------------------------------------------------------------

-- Here's a version that computes the normal as well, for comparison

cylinder' :: FloatE -> R2 -> (R3,R3)
cylinder' h (u,v) = ((c, h*v, s), (c, 0, s))
 where
   theta = 2*pi*u
   c     = cos theta
   s     = sin theta
(cylinder,cylinderN) = splitF2 cylinder'


splitF :: (a -> (b,c)) -> (a -> b, a -> c)
splitF f = (fst . f, snd . f)

splitF2 :: (a -> b -> (c,d)) -> (a -> b -> c, a -> b -> d)
splitF2 f = splitF (splitF . f)


----------------------------------------------------------------------

-- From Namer.hs

-- Types that can be replaced by variables or variable aggregates

class AsVar a where asVar :: a -> Namer (a,TypeEnv)

instance AsVar () where
  asVar _ = return ((),[])

instance (AsVar a, AsVar b) => AsVar (a,b) where
  asVar ~(a,b) = do (a,aenv) <- asVar a
                    (b,benv) <- asVar b
                    return $ ((a,b), aenv++benv)

instance (AsVar a, AsVar b, AsVar c) => AsVar (a,b,c) where
  asVar ~(a,b,c) = do (a,aenv) <- asVar a
                      (b,benv) <- asVar b
                      (c,cenv) <- asVar c
                      return $ ((a,b,c), aenv++benv++cenv)

instance (AsVar a, AsVar b, AsVar c, AsVar d) => AsVar (a,b,c,d) where
  asVar ~(a,b,c,d) = do (a,aenv) <- asVar a
                        (b,benv) <- asVar b
                        (c,cenv) <- asVar c
                        (d,denv) <- asVar d
                        return $ ((a,b,c,d), aenv++benv++cenv++denv)

-- The lazy patterns above are important, because we pass bogus (bottom)
-- values.

unFunctionVM :: AsVar a => (a -> b) -> Namer (b,TypeEnv)
unFunctionVM f = fmap (onFst f) (asVar (error "unFunction: bogus arg"))

unFunction :: AsVar a => (a -> b) -> b
unFunction = fst .
             runNamer allNames .
             unFunctionVM


----------------------------------------------------------------------

trySin2Cos2 :: Try2
trySin2Cos2 e1 e2 = listToMaybe [ one scalar + e1' + e2'
                                | (p1,e1') <- asSum e1
                                , (p2,e2') <- asSum e2
                                , match p1 p2 || match p2 p1
                                ]
 where
   match (Apply Mul [a@(Apply Cos [u]), a']) (Apply Mul [b@(Apply Sin [v]), b']) =
     a ==% a' && b ==% b' && u ==% v
   match _ _ = False

-- TODO: trySin2Cos2 fits a general pattern for AC rules.

-- For testing

tryFactor :: Op1
tryFactor e = fromMaybe e (listToMaybe (tryFactors e))

tryFactors :: Exp -> [Exp]
tryFactors e@(Apply Add [e1,e2]) = factorSum e1 e2
tryFactors e                     = [e]


