-- Some lighting examples.

-- To do: switch to using implicit parameter version (Shader) instead of
-- explicit (EPShader).  Requires working ghc with memo functions and
-- pointer equality.

module Lighting where

import Vertigo
import EPShader


-- Simple directional and point lights

dirLight :: ColorE -> Dir3E -> Light
dirLight col dir = const (col,dir)

pointLight :: ColorE -> Vec3E -> Light
pointLight col lpos p = (col, normalize (lpos - p))

smoothStep :: (FloatE,FloatE) -> FloatE -> FloatE
smoothStep loHi val = t*t*(3-2*t) where t = wrap loHi val

-- To do: add distance-based fall-off and spot lights

stdViewPos = vec3 0.5 1 (-2) :: Vec3E

stdView :: [Light] -> View
stdView lights = (white, stdViewPos, lights)

view1 :: View
view1 = stdView [dirLight white (normalize (- stdViewPos))]

view2 :: Anim Vec3E -> Anim View
view2 pos t = stdView [pointLight red (pos t)]


-- Standard material properties (from rtsl-shaders.in)

ma,md,ms :: ColorE; msh :: FloatE
ma = gray 0.3                          -- ambient
md = gray 0.4                          -- diffuse
ms = gray 0.6                          -- specular
msh = 15                               -- specular exponent

{-
-- Simple shaders with standard material properties
ambDiffStd  = ambDiff ma md
specularStd = specularNH msh
-}
basicStd    = basicRV (ma,md,ms,msh)
