-- Experimenting with linear mappings, for generalized differentiation.

module LinearMap where

import Vector

infixr 0 $# 

-- f is a linear map from a to b
class LMap f a b | f a -> b where ($#) :: f -> a -> b

instance VectorOf s v => LMap s v v where ($#) = (*^)

instance VectorOf s v => LMap v v v where ($#) = (<.>)

-- Examples:
-- l1 = 5 $# (3 :: Vec3E)
-- l2 = (4 :: Vec3E) $# (3 :: Vec3E)


class VectorOf s v => Affine a v | a -> v where
  diffA :: a -> a -> v
  addA  :: a -> v -> a

-- deriv :: LMap lmap a b => (a -> b) -> lmap

data Op      = Add | Mul {- | ... -} | AbsOp Abstr
data Exp     = Apply Op [Exp] | Var String
data Abstr   = Abstr TypeEnv Exp

data Type    = Scalar | Tuple [Type] | Fun Type Type
type TypeEnv = [(String, Type)]

-- Phantom static typing
data E a = E Exp

mkE       = E
unE (E e) = e

abstr :: (E a -> E b) -> E (a -> b)
-- abstr f = mkE (Abstr (mkParams f) ...)
abstr = undefined 

deriv :: (LMap lmap a b, Affine a va, Affine b vb) => E (a -> b) -> lmap
deriv = undefined

-- To go further, we have to use an intensional representation for a->b,
-- so we can see it as function compositions.

