-- "Name structures".  A way to introduce names for parameters based on
-- types.

module NameStruct where

import Types

class AsVar ns dom where
  asVar :: ns -> dom -> (TypeEnv,dom)

unFunction :: AsVar ns a => ns -> (a -> b) -> (TypeEnv,b)
unFunction ns f = onSnd f (asVar ns (error "unFunction: bogus arg"))


-- The lazy patterns above are critical below, because we pass bogus
-- (bottom) values.

instance AsVar () () where
  asVar () ~() = ([],())

instance (AsVar nsA a, AsVar nsB b) => AsVar (nsA,nsB) (a,b) where
  asVar (nsA,nsB) ~(a,b) = (envA++envB, (a',b'))
   where
     (envA,a') = asVar nsA a
     (envB,b') = asVar nsB b

instance (AsVar nsA a, AsVar nsB b, AsVar nsC c)
       => AsVar (nsA,nsB,nsC) (a,b,c) where
  asVar (nsA,nsB,nsC) ~(a,b,c) = (envA++envB++envC, (a',b',c'))
   where
     (envA,a') = asVar nsA a
     (envB,b') = asVar nsB b
     (envC,c') = asVar nsC c

instance (AsVar nsA a, AsVar nsB b, AsVar nsC c, AsVar nsD d)
       => AsVar (nsA,nsB,nsC,nsD) (a,b,c,d) where
  asVar (nsA,nsB,nsC,nsD) ~(a,b,c,d) =
    (envA++envB++envC++envD, (a',b',c',d'))
   where
     (envA,a') = asVar nsA a
     (envB,b') = asVar nsB b
     (envC,c') = asVar nsC c
     (envD,d') = asVar nsD d



-------- Extract names & ranges

-- Named range for float var
data Range = Range Id Float Float Float

class AsRange nrs ns | nrs -> ns where
  asRange :: nrs -> (ns, [Range])

instance AsRange () () where
  asRange () = ((),[])

instance AsRange Range Id where
  asRange range@(Range v _ _ _) = (v,[range])

instance (AsRange rsa nsa, AsRange rsb nsb)
   => AsRange (rsa,rsb) (nsa,nsb) where
  asRange (a,b) = ((nsa,nsb),rsa++rsb)
   where
    (nsa,rsa) = asRange a
    (nsb,rsb) = asRange b

instance (AsRange rsa nsa, AsRange rsb nsb, AsRange rsc nsc)
   => AsRange (rsa,rsb,rsc) (nsa,nsb,nsc) where
  asRange (a,b,c) = ((nsa,nsb,nsc),rsa++rsb++rsc)
   where
    (nsa,rsa) = asRange a
    (nsb,rsb) = asRange b
    (nsc,rsc) = asRange c

instance (AsRange rsa nsa, AsRange rsb nsb, AsRange rsc nsc, AsRange rsd nsd)
   => AsRange (rsa,rsb,rsc,rsd) (nsa,nsb,nsc,nsd) where
  asRange (a,b,c,d) = ((nsa,nsb,nsc,nsd),rsa++rsb++rsc++rsd)
   where
    (nsa,rsa) = asRange a
    (nsb,rsb) = asRange b
    (nsc,rsc) = asRange c
    (nsd,rsd) = asRange d

