-- Name supply monad and supporting types

module Namer where

import Monad

import Types

type Names = [Id]

-- Given a supply of names, come up with an a-typed value, the variables
-- introduced, and the remaining supply.
newtype Namer a = Namer (Names -> (a,Names))

instance Monad Namer where
  return a = Namer (\ names -> (a,names))
  Namer f >>= h = Namer $ \ names ->
                  let (a, names') = f names
                      Namer g = h a
                      (b, names'') = g names'
                  in
                      (b, names'')

instance Functor Namer where
  fmap f m = m >>= return.f


getName :: Namer Id
getName = Namer $ \ (name:names') -> (name,names')

runNamer :: Names -> Namer a -> a
runNamer supply (Namer f) = fst (f supply)

getNames :: Int -> Namer [Id]
getNames n = sequence (replicate n getName)
