-- Counts number of operations in various categories.  Use after CSE.

module NumOps where

import List (intersperse)
import DExp


-- Yields number of operations satisfying given predicate.  Doesn't handle
-- loops, lambdas.
countOps :: (Op -> Bool) -> Exp -> Int

countOps pred = count
 where
   count (LitVec _)                          = 0
   count (Var n t)                           = 0
   count (Let binds body)                    = sum (map (count.snd) binds) + count body
   count (Apply Add [e1, Apply Negate [e2]]) = count e1 + count e2
   count (Apply Add [Apply Negate [e2], e1]) = count e1 + count e2
   count (Apply Mul [e1, Apply Rcp [e2]])    = count e1 + count e2
   count (Apply Mul [Apply Rcp [e2], e1])    = count e1 + count e2
   count (Apply op args)                     =
     (if pred op then 1 else 0) + sum (map count args)


analyzeOps :: Exp -> String
analyzeOps exp = concat $
                 intersperse ", " $
                 map (count exp) categs
 where
   categs = [ ("add", [Add]), ("multiply",[Mul]), ("trig",[Sin,Cos]) ]
            -- [("arithmetic",arith),("transendental",transend)]
   count exp (name,ops) = show (countOps (flip elem ops) exp) ++ " " ++ name

   arith    = [Add,Max,Min,Mul,Sge,Slt,Rcp,Dp3,Dp4,Negate]
   transend = [Sin,Cos,Log,Exp]
