-- Procedural geometry

module ProcGeom where

import Vertigo
import EPShader (SurfPt)

-- Procedural parametric surface.  Maps two-space to point, normal and
-- color.
type Surf = Vec2E -> SurfPt

xyPlane :: ColorE -> Surf
xyPlane col p = SurfPt (vec3 x y 0) (vec3 0 0 1) col where (x,y) = coords p

-- Cylinder from radius, height and color
cylinder :: FloatE -> ColorE -> Surf
cylinder h col p = SurfPt (vec3 c (h*v) s) (vec3 c 0 s) col
 where
   (u,v) = coords p
   theta = 2*pi*u
   c     = cos theta
   s     = sin theta

-- Sphere from radius and color
sphere :: ColorE -> Surf
sphere col p = SurfPt (vec3 spct cp spst)
                      (vec3 spct cp spst)
                      col
 where
   (u,v) = coords p
   theta = pi*u
   phi   = 2*pi*v
   cp    = cos phi
   spct  = sin phi * cos theta
   spst  = sin phi * sin theta

-- TODO: I think it should be theta=2*u & phi=v, but the above works
-- instead.  Investigate.
