-- Some experiments in rewrite rule specification.

module Rewrite where

import List (intersperse)

import Types
import qualified DExp as D
import Namer
import TExp

class EFun f b where
  toBase :: f -> Namer (TypeEnv,b)

instance EFun (Exp a) (Exp a) where
  toBase e = return ([],e)

instance (AsVector a, EFun f b) => EFun (Exp a -> f) b where
  toBase g = do va <- getName
                (env,e) <- toBase (g (var va))
                return $ ((va, expType (argument g)):env,e)

argument :: (a -> b) -> a
argument = error "argument undefined"

data Rule = Rule TypeEnv D.Exp D.Exp

instance Show Rule where
  show (Rule env lhs rhs) = "for all " ++ showEnv env ++
                            "\n   " ++ sh lhs ++ " ==> " ++ sh rhs
   where
     showEnv env = concat (intersperse ", " (map showBinding env))
     showBinding (var,ty) = var ++ " :: " ++ show ty
     sh dexp = show (E dexp :: FloatE)

data RuleBody = RuleBody D.Exp D.Exp

instance EFun RuleBody RuleBody where
  toBase body = return ([],body)

infix 1 ==> -- , `condR`
E lhs ==> E rhs = RuleBody lhs rhs

rule :: EFun f RuleBody => f -> Rule
rule f = Rule env lhs rhs
 where
   (env, RuleBody lhs rhs) =
     runNamer allIds (toBase f)

r1 = rule $ \ (a :: FloatE) b ->
     sqrt a * sqrt b ==> sqrt (a*b)

-- Note that we had to monomorphize r1.  That's a shame.  Is it avoidable?


-- We can't do this next one, because y & z are floats, not expressions.
-- I think I just need a bit richer type system.  Think on it.  This rule
-- is sort of peculiar.  Like a meta-rule.
{-
r2 = rule $ \ a y z ->
     a + constE y <* constE z  ==>  a <* constE (z-y)
-}

{-

someRules =
  [ 
    rules $ \ a y z ->
      [ a + lit y < lit z  ==>  a < lit (z-y)
      , lit z < a + lit y  ==>  lit (z-y) < a
      ]

  , rule $ \ a x b ->
    a * lit x < b  ==>  [ a < b / lit x `condR` x > 0
                        , a > b / lit x
                        ]
  ]

-}