-- Sample statically typed expressions

module SampleExp where

import TExp
import SwizzleFuns
import Vector
import qualified SmartDExp as Sm

-- Some variables 1D, 2D, 3D 
v1@[q,r,s] = map (var . (:[])) "qrs" :: [Vec1E]
v2@[a,b,c] = map (var . (:[])) "abc" :: [Vec2E]
v3@[e,f,g] = map (var . (:[])) "efg" :: [Vec3E]

vertexVars = map unE v3 ++ map unE v2 ++ map unE v1

-- Test these ones with testC
e1 = - a
e2 = a+b                                --  a+b
e3 = 3-a                                --  3-a
e4 = a*e3                               --  a*(3-a)
e5 = e4 + e4                            --  let b = a*(3-a) in b+b
e6 = e2*e5                              --  (a+b)*e5
e7 = e2*e3                              --  (a+b)*(3-a)
e8 = negate e1                          --  -(-a)
e9 = negate e2                          --  -(a+b)

e10 = a <* (vec2 3 4 :: Vec2E)
e11 = a >* b
e12 = notE e10
e13 = e11 &&* e12
e14 = e11 ||* e12

e15 = sin q
e16 = e<.>e + 1
e17 = sin q + cos r
e18 = sin r + cos r

e19 = e7+e1                             -- mad
e20 = -e4
e21 = - min e f

-- Test this one with testOut
e22 = output [oPos e7, oSpecular e5, oDiffuse e3]

e23 = e21 + swizYXY e2    --  -(min e f) +(a+b).yxy
e24 = - (swizYX (a `max` b))
e25 = swizYX (- (a `max` b))
e26 = swizYX (swizYX e2)

e27 = let u = swizYX e2 in u + u

e28 = wrap (-pi,pi) q
e29 = frac q

-- Try unFunction on these functions.

e30 u = v + v where v = u * (3-u) :: FloatE
e31 (s,t) = v + v where v = s * (3-t) :: FloatE

-- Staged functions.  Apply "test" to these ones.

e33 c     (s,t)     = v + v where v = s * (c-t) :: Vec3E
e34 (c,d) (r,(s,t)) = c*r + d*s - 3*t :: Vec2E

-- Trig identities
e35 = sin q ^ 2 + cos q ^ 2


-- Automatic vectorization
e40 = vec3 (q+r) (r+s) (s+q)   -- vec3 q r s + vec3 r s q
e41 = vec3 (-q) (-r) (-s)
e45 = fromCoords (coords e `cross3` coords f)
