module TestDeriv where

import TExp
import Derivative
import DExp (Apply,Op(..))
import qualified NumOps
import SwizzleFuns

varNames = map (:[]) "abcdxyz"
[a,b,c,d,x,y,z] = map var varNames :: [FloatE]

csev :: CSE a => a -> a
csev = cse varNames

xy = (x,y)

dx :: Derivative a => a -> a
dx = derivative "x" 

dxy :: Derivative a => a -> (a,a)
dxy = derivPair ("x","y")

dv1 = dx (3*x*y)
dv2 = dxy (3*x*y)

dv3 = dx (3 ** x)

dv4 = dxy (sin (x+y) ** y)

dv4' = csev (u+v) where (u,v) = dv4

dv4a  = fst dv4
dv4a' = csev dv4a

dv4b  = snd dv4
dv4b' = csev dv4b

pyrN = 3 :: Int

-- pyramid :: ExpType a => Int -> (Exp a -> Exp a -> Exp a) -> Exp a
pyramid n phi = head $ (!! (n-1)) $
                iterate (onPairs phi) $
                map (var . (:[])) $
                take n ['a' .. 'z']

onPairs :: (a -> a -> a) -> [a] -> [a]
onPairs f xs = zipWith f xs (tail xs)
                
pyrVars = map (:[]) (take pyrN ['a'..'z'])

csez :: CSE a => a -> a
csez = cse pyrVars

dv13 = pyramid pyrN (\ u v -> cos u * sin v) :: FloatE
dv14 = vec [derivative var dv13 | var <- pyrVars]

dv13' = csez dv13
dv14' = csez dv14

dv15  = derivative "a" dv13
dv15' = csez dv15

-- To do next: compute Hessian, i.e., the matrix of mixed second partial
-- derivatives.  [ derivative u (derivative v e) | u <- vars, v <- vars ]

vec :: [FloatE] -> Exp ()
vec = mkE . Apply MkVec . map unE

-- Examples from Brian's automatic differentiation implementation

bg3 = vec3 x0 x1 x2
      where
        v0   = cos(a)
        v1   = sin(b)
        v2   = v0*v1
        v3   = cos(v2)
        v4   = cos(b)
        v5   = sin(c)
        v6   = v4*v5
        v7   = sin(v6)
        v8   = v3*v7
        vr9  = 1
        vr10 = vr9*v3
        vr11 = cos(v6)*vr10
        vr12 = vr11*v4
        x2   = cos(c)*vr12
        vr13 = vr11*v5
        x1   = -v1*vr13+v4*vr16
        vr14 = vr9*v7
        vr15 = -sin(v2)*vr14
        vr16 = vr15*v0
        vr17 = vr15*v1
        x0   = -sin(a)*vr17

bg3' = csez bg3


bg4 = vec4 x0 x1 x2 x3
      where
        v0    = cos(a)
        v1    = sin(b)
        v2    = v0*v1
        v3    = cos(v2)
        v4    = cos(b)
        v5    = sin(c)
        v6    = v4*v5
        v7    = sin(v6)
        v8    = v3*v7
        v9    = cos(v8)
        v10   = cos(v6)
        v11   = cos(c)
        v12   = sin(d)
        v13   = v11*v12
        v14   = sin(v13)
        v15   = v10*v14
        v16   = sin(v15)
        v17   = v9*v16
        vr18  = 1
        vr19  = vr18*v9
        vr20  = cos(v15)*vr19
        vr21  = vr20*v10
        vr22  = cos(v13)*vr21
        vr23  = vr22*v11
        x3    = cos(d)*vr23
        vr24  = vr22*v12
        x2    = -v5*vr24 + v11*vr30
        vr25  = vr20*v14
        vr29  = -v7*vr25
        vr26  = vr18*v16
        vr27  = -sin(v8)*vr26
        vr28  = vr27*v3
        vr29' = vr29 + v10*vr28
        vr30  = vr29'*v4
        vr31  = vr29'*v5
        x1    = -v1*vr31 + v4*vr34
        vr32  = vr27*v7
        vr33  = -sin(v2)*vr32
        vr34  = vr33*v0
        vr35  = vr33*v1
        x0    = -sin(a)*vr35

bg4' = csez bg4
